"""
Copied from Sphinx plugins for Django documentation.-----------------------------------------------
"""

from importlib import import_module
from inspect import getsource

from docutils import nodes
from docutils.parsers.rst import Directive, directives
from sphinx import addnodes
from sphinx.domains.std import Cmdoption
from sphinx.util.nodes import set_source_info


def setup(app):
    app.add_crossref_type(
        directivename="setting",
        rolename="setting",
        indextemplate="pair: %s; setting",
    )
    app.add_crossref_type(
        directivename="templatetag",
        rolename="ttag",
        indextemplate="pair: %s; template tag"
    )
    app.add_crossref_type(
        directivename="templatefilter",
        rolename="tfilter",
        indextemplate="pair: %s; template filter"
    )
    app.add_crossref_type(
        directivename="fieldlookup",
        rolename="lookup",
        indextemplate="pair: %s; field lookup type",
    )
    app.add_directive('django-admin-option', Cmdoption)
    app.add_config_value('django_next_version', '0.0', True)

    # register the snippet directive
    app.add_directive('snippet', SnippetWithFilename)
    # register a node for snippet directive so that the xml parser
    # knows how to handle the enter/exit parsing event
    app.add_node(snippet_with_filename,
                 html=(visit_snippet, depart_snippet_literal),
                 man=(visit_snippet_literal, depart_snippet_literal),
                 text=(visit_snippet_literal, depart_snippet_literal),
                 texinfo=(visit_snippet_literal, depart_snippet_literal))

    app.add_directive('pprint', PrettyPrintIterable)

    return {'parallel_read_safe': True}


class snippet_with_filename(nodes.literal_block):
    """
    Subclass the literal_block to override the visit/depart event handlers
    """
    pass


def visit_snippet_literal(self, node):
    """
    default literal block handler
    """
    self.visit_literal_block(node)


def depart_snippet_literal(self, node):
    """
    default literal block handler
    """
    self.depart_literal_block(node)


def visit_snippet(self, node):
    """
    HTML document generator visit handler
    """
    lang = node["language"]
    linenos = node.get('linenos', False)
    fname = node['filename']
    highlight_args = node.get('highlight_args', {})
    if 'language' in node:
        # code-block directives
        lang = node['language']
        highlight_args['force'] = True
    if 'linenos' in node:
        linenos = node['linenos']

    def warner(msg):
        self.builder.warn(msg, (self.builder.current_docname, node.line))

    highlighted = self.highlighter.highlight_block(node.rawsource, lang,
                                                   warn=warner,
                                                   linenos=linenos,
                                                   **highlight_args)
    starttag = self.starttag(node, 'div', suffix='',
                             CLASS='highlight-%s snippet' % lang)
    self.body.append(starttag)
    self.body.append('<div class="snippet-filename">%s</div>\n''' % (fname,))
    self.body.append(highlighted)
    self.body.append('</div>\n')
    raise nodes.SkipNode


class SnippetWithFilename(Directive):
    """
    The 'snippet' directive that allows to add the filename (optional)
    of a code snippet in the document. This is modeled after CodeBlock.
    """
    has_content = True
    optional_arguments = 1
    option_spec = {'filename': directives.unchanged_required}

    def run(self):
        code = '\n'.join(self.content)

        literal = snippet_with_filename(code, code)
        if self.arguments:
            literal['language'] = self.arguments[0]
        literal['filename'] = self.options['filename']
        set_source_info(self, literal)
        return [literal]


def parse_django_admin_node(env, sig, signode):
    command = sig.split(' ')[0]
    env.ref_context['std:program'] = command
    title = "django-admin %s" % sig
    signode += addnodes.desc_name(title, title)
    return command


#----------------Human readable iterables in Sphinx documentation--------
# https://stackoverflow.com/a/62253904/3437454

class PrettyPrintIterable(Directive):
    required_arguments = 1

    def run(self):
        def _get_iter_source(src, varname):
            # 1. identifies target iterable by variable name, (cannot be spaced)
            # 2. determines iter source code start & end by tracking brackets
            # 3. returns source code between found start & end
            start = end = None
            open_brackets = closed_brackets = 0
            for i, line in enumerate(src):
                if line.startswith(varname):
                    if start is None:
                        start = i
                if start is not None:
                    open_brackets   += sum(line.count(b) for b in "([{")
                    closed_brackets += sum(line.count(b) for b in ")]}")

                if open_brackets > 0 and (open_brackets - closed_brackets == 0):
                    end = i + 1
                    break
            return '\n'.join(src[start:end])

        module_path, member_name = self.arguments[0].rsplit('.', 1)
        src = getsource(import_module(module_path)).split('\n')
        code = _get_iter_source(src, member_name)

        literal = nodes.literal_block(code, code)
        literal['language'] = 'python'

        return [addnodes.desc_content('', literal)]



from django.conf import settings
from django.core.files.storage import default_storage
from django.db import models
from django.utils.translation import gettext_lazy as _


class BuiltInGalleryImage(models.Model):
    image = models.ImageField(
        upload_to="images", storage=default_storage, verbose_name=_("Image"))
    creator = models.ForeignKey(
        settings.AUTH_USER_MODEL, null=False, blank=False,
        verbose_name=_('Creator'), on_delete=models.CASCADE)

    # This snippet will not run because we already have the ImageField named 'image'
    # @classmethod
    # def get_image_field(cls):
    #     return cls._meta.get_field("image")



from django.contrib.auth.decorators import login_required
from django.urls import path

from galleryfield import image_views

# The default view names also follow the form of "app_label-model_name-upload".

urlpatterns = [
    path('upload/',
         login_required(image_views.BuiltInImageCreateView.as_view()),
         name="galleryfield-builtingalleryimage-upload"),
    path('fetch/',
         login_required((image_views.BuiltInImageListView.as_view())),
         name="galleryfield-builtingalleryimage-fetch"),
    path('crop/<int:pk>',
         login_required(image_views.BuiltInImageCropView.as_view()),
         name="galleryfield-builtingalleryimage-crop"),
]



from django.core.exceptions import PermissionDenied
from django.views.generic import CreateView, UpdateView
from django.views.generic.list import BaseListView

from galleryfield.mixins import (BaseCreateMixin, BaseCropViewMixin,
                                 BaseListViewMixin)


class ImageCreateView(BaseCreateMixin, CreateView):
    """
    The Class-based view handling the saving of uploaded image.

    .. attribute:: target_model

       |view_target_model|

    .. attribute:: crop_url_name

       |view_crop_url_name|

    .. attribute:: disable_server_side_crop

       |view_disable_server_side_crop|

    .. automethod:: form_valid
    """
    def form_valid(self, form):
        """User should override this method to save the object,
        for example, image model usually has a non-null user field,
        that should be handled here.

        See `the-save-method <https://docs.djangoproject.com/en/dev/topics/forms/modelforms/#the-save-method>`_
        for detail.

        See :class:`galleryfield.image_views.BuiltInImageCreateView` for example.
        """  # noqa
        self.object.save()
        return super().form_valid(form)


class ImageListView(BaseListViewMixin, BaseListView):
    """
    The Class-based view for fetching the existing images of the gallery instance.

    .. attribute:: target_model

       |view_target_model|

    .. attribute:: crop_url_name

       |view_crop_url_name|

    .. attribute:: disable_server_side_crop

       |view_disable_server_side_crop|

    .. automethod:: get_queryset
    """
    def get_queryset(self):
        """
        User need to override this method to do some basic filter in terms of
        who can see which images.

        :return: A Queryset
        """
        return super().get_queryset()


class ImageCropView(BaseCropViewMixin, UpdateView):
    """
    The Class-based view handling server side cropping of an image model
    instance.

    .. attribute:: target_model

       |view_target_model|

    .. attribute:: crop_url_name

       |view_crop_url_name|

    .. attribute:: disable_server_side_crop

       |view_disable_server_side_crop|

    .. automethod:: form_valid
    """
    def form_valid(self, form):
        """Save the object from the form. This method should only be overridden
        when the model contains dynamic fields which need to be updated. For
        example, when there is a  ``DateTimeField`` which records the updated
        datetime of the image.
        """
        self.object = form.save()
        return super().form_valid(form)


class BuiltInImageCreateView(ImageCreateView):
    target_model = "galleryfield.BuiltInGalleryImage"
    crop_url_name = "galleryfield-builtingalleryimage-crop"  # Can be omitted
    disable_server_side_crop = False

    def form_valid(self, form):
        self.object = form.save(commit=False)
        self.object.creator = self.request.user
        self.object.save()
        return super().form_valid(form)


class BuiltInImageListView(ImageListView):
    target_model = "galleryfield.BuiltInGalleryImage"
    crop_url_name = "galleryfield-builtingalleryimage-crop"  # Can be omitted
    disable_server_side_crop = False

    def get_queryset(self):
        queryset = super().get_queryset()
        if not self.request.user.is_superuser:
            queryset = queryset.filter(creator=self.request.user)
        return queryset


class BuiltInImageCropView(ImageCropView):
    target_model = "galleryfield.BuiltInGalleryImage"
    crop_url_name = "galleryfield-builtingalleryimage-crop"  # Can be commented
    disable_server_side_crop = False

    def get_object(self, queryset=None):
        obj = super().get_object(queryset=None)
        if not self.request.user.is_superuser and obj.creator != self.request.user:
            raise PermissionDenied("May not crop other people's image")
        return obj



# Generated by Django 3.2.5 on 2021-08-02 05:40

from django.db import migrations, models

import galleryfield.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='DemoGalleryForTest',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('images', galleryfield.fields.GalleryField(target_model='tests.FakeValidImageModel')),
            ],
        ),
        migrations.CreateModel(
            name='FakeInvalidImageModel1',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('image', models.CharField(max_length=250)),
            ],
        ),
        migrations.CreateModel(
            name='FakeInvalidImageModel2',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('photo', models.ImageField(upload_to='')),
            ],
        ),
        migrations.CreateModel(
            name='FakeInvalidImageModel3',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('photo', models.CharField(max_length=250)),
            ],
        ),
        migrations.CreateModel(
            name='FakeInvalidImageModel4',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('photo', models.CharField(max_length=250)),
            ],
        ),
        migrations.CreateModel(
            name='FakeInvalidImageModel5',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('photo', models.ImageField(upload_to='')),
            ],
        ),
        migrations.CreateModel(
            name='FakeValidImageModel',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('image', models.ImageField(upload_to='')),
            ],
        ),
    ]
